@@ -2,6 +2,7 @@ use cdx::{Error, Result};
 use std::env;
 
 mod args;
+mod binsearch_main;
 mod cat_main;
 mod cdxmain;
 mod cgrep_main;
@@ -44,10 +45,14 @@ fn inner_main(mut args: Vec<String>) -> Result<()> {
         println!("USAGE : cdx <command> [options...]");
         println!("Commands are :");
         for x in cdxmain::MAINLIST {
-            println!("{} :\t{}", x.name, x.help);
+            println!("{:8} : {}", x.name, x.help);
         }
         return Ok(());
     }
+    if args[1] == "version" || args[1] == "--version" {
+        println!("cdx version {}", args::version());
+        return Ok(());
+    }
     for x in cdxmain::MAINLIST {
         if args[1] == x.name {
             let arg1 = args.remove(1);
@@ -58,7 +63,7 @@ fn inner_main(mut args: Vec<String>) -> Result<()> {
     }
     eprintln!("Valid subcommands are :");
     for x in cdxmain::MAINLIST {
-        eprintln!("{} :\t{}", x.name, x.help);
+        eprintln!("{:8} : {}", x.name, x.help);
     }
     Err(Error::Silent)
 }
