@@ -1,47 +1,12 @@
 //! Test if a line matches some criterea
 
 use crate::column::NamedCol;
-use crate::{bglob, err, sglob, Error, Reader, Result, TextLine};
+use crate::text::{Case, Text};
+use crate::{err, Error, Reader, Result, TextLine};
 use memchr::memmem::find;
 use std::collections::HashSet;
 use std::fmt;
 
-/// compare for equality, ignoring case.
-fn equal_nocase(a: &str, b: &str) -> bool {
-    for ch in a
-        .chars()
-        .flat_map(char::to_lowercase)
-        .zip(b.chars().flat_map(char::to_lowercase))
-    {
-        if ch.0 != ch.1 {
-            return false;
-        }
-    }
-    a.len() == b.len()
-}
-/*
-/// compare for equality, ignoring case, 'a' is already lowercase
-fn equal_nocase_a(a: &str, b: &str) -> bool {
-    for ch in a
-        .chars()
-        .zip(b.chars().flat_map(char::to_lowercase))
-    {
-    if ch.0 != ch.1 {
-        return false;
-    }
-    }
-    a.len() == b.len()
-}
-*/
-/*
-fn assign_lower(dst: &mut String, src : &str) {
-    dst.clear();
-    dst.extend(src.chars().flat_map(char::to_lowercase));
-}
- */
-
-// self.len() == other.len() && iter::zip(self, other).all(|(a, b)| a.eq_ignore_ascii_case(b))
-
 /// Match against whole line
 pub trait LineCheck {
     /// is the line ok?
@@ -78,6 +43,16 @@ pub enum CheckType {
     /// if pattern is one number, a string a least that long
     /// if two numbers, length between the two numbers, inclusive
     Length,
+    /// length zero
+    Empty,
+    /// all whitespace
+    Blank,
+    /// all whitespace, up to a #
+    HashCmt,
+    /// all whitespace, up to a //
+    SlashCmt,
+    /// all whitespace, up to a literal pattern
+    Comment,
     // XmlAttr
     // DelimSuffix
     // DelimPrefix
@@ -86,10 +61,11 @@ pub enum CheckType {
 }
 impl Default for CheckType {
     fn default() -> Self {
-        CheckType::Regex
+        Self::Regex
     }
 }
 
+/// pattern is prefix of string
 #[derive(Debug, Clone)]
 struct PrefixCheck {
     data: String,
@@ -110,6 +86,105 @@ impl BufCheck for PrefixCheck {
     }
 }
 
+#[derive(Debug, Clone, Copy)]
+/// check all whitespace
+struct BlankCheck;
+
+fn is_blank(data: &[u8]) -> bool {
+    for x in data {
+        if *x > b' ' {
+            return false;
+        }
+    }
+    true
+}
+
+impl BufCheck for BlankCheck {
+    fn scheck(&self, buff: &str) -> bool {
+        is_blank(buff.as_bytes()) // FIXME unicode blanks?
+    }
+    fn ucheck(&self, buff: &[u8]) -> bool {
+        is_blank(buff)
+    }
+}
+
+#[derive(Debug, Clone)]
+/// check all whitespace, followed by literal
+struct CmtCheck {
+    tag: String,
+}
+impl CmtCheck {
+    fn new(tag: &str) -> Self {
+        Self {
+            tag: tag.to_string(),
+        }
+    }
+    /*
+        Replaced by Text::is_cmt, which might be slower
+        fn is_ucmt(&self, data: &[u8]) -> bool {
+            let mut pos: usize = 0;
+            let mut found = false;
+        let tag : &[u8] = self.tag.as_bytes();
+            for x in data {
+                if found {
+                    if pos >= tag.len() {
+                        return true;
+                    }
+                    if tag[pos] != *x {
+                        return false;
+                    }
+                    pos += 1;
+                } else if *x > b' ' {
+                    if tag.is_empty() {
+                        return false;
+                    }
+                    found = true;
+                    if tag[0] != *x {
+                        return false;
+                    }
+                    pos = 1;
+                }
+            }
+            true
+        }
+
+        fn is_scmt(&self, data: &str) -> bool {
+            let mut found = false;
+        let mut tag : &str = &self.tag;
+            for x in data.chars() {
+                if found {
+            if tag.is_empty() {
+                        return true;
+                    }
+                    if tag.first() != x {
+                        return false;
+                    }
+            tag = tag.skip_first();
+                } else if x > ' ' && !x.is_whitespace() {
+                    if tag.is_empty() {
+                        return false;
+                    }
+                    found = true;
+                    if tag.first() != x {
+                        return false;
+                    }
+            tag = tag.skip_first();
+                }
+            }
+            true
+        }
+    */
+}
+
+impl BufCheck for CmtCheck {
+    fn scheck(&self, buff: &str) -> bool {
+        buff.is_cmt(&self.tag)
+    }
+    fn ucheck(&self, buff: &[u8]) -> bool {
+        buff.is_cmt(self.tag.as_bytes())
+    }
+}
+
 fn str_to_lower(data: &str, unicode: bool) -> String {
     if unicode {
         data.to_lowercase()
@@ -119,6 +194,7 @@ fn str_to_lower(data: &str, unicode: bool) -> String {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is prefix of string, case insensitive
 struct PrefixCheckC {
     data: String,
 }
@@ -206,6 +282,7 @@ impl BufCheck for PrefixCheckC {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is suffix of string
 struct SuffixCheck {
     data: String,
 }
@@ -226,6 +303,7 @@ impl BufCheck for SuffixCheck {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is suffix of string, case insensitive
 struct SuffixCheckC {
     data: String,
 }
@@ -246,6 +324,7 @@ impl BufCheck for SuffixCheckC {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is substring of string, case insensitive
 struct InfixCheckC {
     data: String,
 }
@@ -266,6 +345,7 @@ impl BufCheck for InfixCheckC {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is substring of string
 struct InfixCheck {
     needle: Vec<u8>,
 }
@@ -286,6 +366,7 @@ impl BufCheck for InfixCheck {
 }
 
 #[derive(Debug, Clone)]
+/// bytes regex
 struct RegexCheckB {
     data: regex::bytes::Regex,
 }
@@ -306,6 +387,7 @@ impl BufCheck for RegexCheckB {
 }
 
 #[derive(Debug, Clone)]
+/// utf8 regex
 struct RegexCheckS {
     data: regex::Regex,
 }
@@ -326,6 +408,7 @@ impl BufCheck for RegexCheckS {
 }
 
 #[derive(Debug, Clone)]
+/// string exactly matches pattern
 struct ExactCheck {
     data: Vec<u8>,
 }
@@ -365,6 +448,7 @@ fn load_hashset(data: &mut HashSet<Vec<u8>>, fname: &str) -> Result<()> {
 }
 
 #[derive(Debug, Clone)]
+/// pattern is file name. String exactly matches one line of file.
 struct FileExactCheck {
     data: HashSet<Vec<u8>>,
 }
@@ -392,11 +476,16 @@ fn load_hashset_c(data: &mut HashSet<Vec<u8>>, fname: &str, unicode: bool) -> Re
         return Ok(());
     }
     loop {
-        let line = &f.curr().line;
+        let mut line: &[u8] = &f.curr().line;
         if line.len() > 1 {
+            if line.last().unwrap() == &b'\n' {
+                line = &line[..line.len() - 1];
+            }
             if unicode {
+                data.insert(String::from_utf8(line.to_vec())?.new_lower().into_bytes());
+            // PERF - 2 allocations
             } else {
-                data.insert(line[0..line.len() - 1].to_ascii_lowercase());
+                data.insert(line.new_lower());
             }
         }
         if f.getline()? {
@@ -407,6 +496,7 @@ fn load_hashset_c(data: &mut HashSet<Vec<u8>>, fname: &str, unicode: bool) -> Re
 }
 
 #[derive(Debug, Clone)]
+/// pattern is file name. String exactly matches one line of file, case insensitive
 struct FileExactCheckC {
     data: HashSet<Vec<u8>>,
 }
@@ -419,20 +509,21 @@ impl FileExactCheckC {
 }
 impl BufCheck for FileExactCheckC {
     fn scheck(&self, buff: &str) -> bool {
-        self.data.contains(buff.to_lowercase().as_bytes()) // PERF allocation
+        self.data.contains(&buff.new_lower().into_bytes()) // PERF allocation
     }
     fn ucheck(&self, buff: &[u8]) -> bool {
-        self.data.contains(&buff.to_ascii_lowercase()) // PERF allocation
+        self.data.contains(&buff.new_lower()) // PERF allocation
     }
 }
 
 #[derive(Debug, Clone)]
+/// shell style wildcard match
 struct GlobCheck {
     data: String,
-    ic: bool,
+    ic: Case,
 }
 impl GlobCheck {
-    fn new(data: &str, ic: bool) -> Self {
+    fn new(data: &str, ic: Case) -> Self {
         Self {
             data: data.to_string(),
             ic,
@@ -441,45 +532,52 @@ impl GlobCheck {
 }
 impl BufCheck for GlobCheck {
     fn scheck(&self, buff: &str) -> bool {
-        sglob(&self.data, buff, self.ic)
+        buff.glob(&self.data, self.ic)
     }
     fn ucheck(&self, buff: &[u8]) -> bool {
-        bglob(self.data.as_bytes(), buff, self.ic)
+        buff.glob(self.data.as_bytes(), self.ic)
     }
 }
 
 #[derive(Debug, Clone)]
+/// string exactly matches pattern, case insensitive
 struct ExactCheckC {
     data: String,
 }
-// PERF - should lowercase data in constructor, use simpler comparitor
+
 impl ExactCheckC {
-    fn new(data: &str) -> Self {
-        Self {
-            data: data.to_string(),
-        }
+    fn new(data: &str, unicode: bool) -> Result<Self> {
+        Ok(Self {
+            data: if unicode {
+                data.new_lower()
+            } else {
+                String::from_utf8(data.as_bytes().new_lower())?
+            },
+        })
     }
 }
 impl BufCheck for ExactCheckC {
     fn scheck(&self, buff: &str) -> bool {
-        equal_nocase(&self.data, buff)
+        buff.equal_insens_quick(&self.data)
     }
     fn ucheck(&self, buff: &[u8]) -> bool {
-        self.data.as_bytes().eq_ignore_ascii_case(buff)
+        buff.equal_insens_quick(self.data.as_bytes())
     }
 }
 
-#[derive(Debug, Clone, Default)]
+#[derive(Debug, Clone, Default, Copy)]
+/// check length
 struct LengthCheck {
     min: usize,
     max: Option<usize>,
 }
 impl LengthCheck {
+    /// new
     fn new(data: &str) -> Result<Self> {
         if data.is_empty() {
             return err!("Length spec can't be empty");
         }
-        let mut val = LengthCheck::default();
+        let mut val = Self::default();
         for (n, x) in data.split(',').enumerate() {
             if n == 0 {
                 val.min = x.parse::<usize>()?;
@@ -491,6 +589,12 @@ impl LengthCheck {
         }
         Ok(val)
     }
+    /*
+        /// new from actual numbers
+        pub fn with_sizes(min : usize, max : Option<usize>) -> Self {
+        Self {min, max}
+        }
+    */
 }
 
 impl BufCheck for LengthCheck {
@@ -516,18 +620,26 @@ impl BufCheck for LengthCheck {
     }
 }
 
+/// return a new BufCheck based on spec
+pub fn make_match(spec: &str) -> Result<Box<dyn BufCheck>> {
+    if let Some((x, y)) = spec.split_once(',') {
+        CheckSpec::new(x)?.make_box(y)
+    } else {
+        CheckSpec::new(spec)?.make_box("")
+    }
+}
+
 /// Spec for BufCheck
-#[derive(Debug, Copy, Clone, Default)]
+#[derive(Debug, Clone, Default)]
 pub struct CheckSpec {
     /// general type
     ctype: CheckType,
     /// true for unicode and &str, false for bytes an &[u8]
     string: bool,
-
     // u8 to unicode error, true for error, false for lossy
     // strict : bool,
-    /// true for case sensitive, false for case insensitive
-    case_insensitive: bool,
+    /// should match be case sensitive
+    case: Case,
     /// true to invert the match
     negate: bool,
 }
@@ -539,8 +651,8 @@ impl CheckSpec {
     /// C for case-insensitive vs default of case-sensitive
     // F for fail-on-utf8-error vs default of lossy
     /// type which is one of : regex, exact, prefix, suffix, substring, fileexact, glob, length. Default is regex.
-    fn new(spec: &str) -> Result<Self> {
-        let mut c = CheckSpec::default();
+    pub fn new(spec: &str) -> Result<Self> {
+        let mut c = Self::default();
         for x in spec.split('.') {
             if x.eq_ignore_ascii_case("") {
             } else if x.eq_ignore_ascii_case("S") {
@@ -548,7 +660,7 @@ impl CheckSpec {
             } else if x.eq_ignore_ascii_case("N") {
                 c.negate = true;
             } else if x.eq_ignore_ascii_case("C") {
-                c.case_insensitive = true;
+                c.case = Case::Insens;
             } else if x.eq_ignore_ascii_case("regex") {
                 c.ctype = CheckType::Regex;
             } else if x.eq_ignore_ascii_case("exact") {
@@ -568,6 +680,16 @@ impl CheckSpec {
                 c.ctype = CheckType::Glob;
             } else if x.eq_ignore_ascii_case("length") {
                 c.ctype = CheckType::Length;
+            } else if x.eq_ignore_ascii_case("blank") {
+                c.ctype = CheckType::Blank;
+            } else if x.eq_ignore_ascii_case("empty") {
+                c.ctype = CheckType::Empty;
+            } else if x.eq_ignore_ascii_case("hash") {
+                c.ctype = CheckType::HashCmt;
+            } else if x.eq_ignore_ascii_case("slash") {
+                c.ctype = CheckType::SlashCmt;
+            } else if x.eq_ignore_ascii_case("comment") {
+                c.ctype = CheckType::Comment;
             } else {
                 return err!(
                     "Invalid Checker Item {}. Should be period delimited list :\n\
@@ -575,7 +697,7 @@ impl CheckSpec {
 			     'S' for string vs default of bytes\n\
 			     'N' for negate vs default of normal\n\
 			     'C' for case-insensitive vs default of case-sensitive\n\
-			     plus optionally one of these, to overrie the default 'regex'\n\
+			     plus optionally one of these, to override the default 'regex'\n\
 			     'regex' field must match pattern as per the regex crate\n\
 			     'exact' field must match pattern exactly\n\
 			     'prefix' pattern must be a prefix of the field\n\
@@ -602,42 +724,47 @@ impl CheckSpec {
                 }
             }
             CheckType::FileExact => {
-                if self.case_insensitive {
+                if self.case == Case::Insens {
                     Box::new(FileExactCheckC::new(pattern, self.string)?)
                 } else {
                     Box::new(FileExactCheck::new(pattern)?)
                 }
             }
             CheckType::Exact => {
-                if self.case_insensitive {
-                    Box::new(ExactCheckC::new(pattern))
+                if self.case == Case::Insens {
+                    Box::new(ExactCheckC::new(pattern, self.string)?)
                 } else {
                     Box::new(ExactCheck::new(pattern))
                 }
             }
             CheckType::Length => Box::new(LengthCheck::new(pattern)?),
             CheckType::Prefix => {
-                if self.case_insensitive {
+                if self.case == Case::Insens {
                     Box::new(PrefixCheckC::new(pattern, self.string))
                 } else {
                     Box::new(PrefixCheck::new(pattern))
                 }
             }
             CheckType::Suffix => {
-                if self.case_insensitive {
+                if self.case == Case::Insens {
                     Box::new(SuffixCheckC::new(pattern, self.string))
                 } else {
                     Box::new(SuffixCheck::new(pattern))
                 }
             }
             CheckType::Substring => {
-                if self.case_insensitive {
+                if self.case == Case::Insens {
                     Box::new(InfixCheckC::new(pattern, self.string))
                 } else {
                     Box::new(InfixCheck::new(pattern))
                 }
             }
-            CheckType::Glob => Box::new(GlobCheck::new(pattern, self.case_insensitive)),
+            CheckType::Glob => Box::new(GlobCheck::new(pattern, self.case)),
+            CheckType::Empty => Box::new(LengthCheck::new("0,0")?),
+            CheckType::Blank => Box::new(CmtCheck::new("")),
+            CheckType::Comment => Box::new(CmtCheck::new(pattern)),
+            CheckType::HashCmt => Box::new(CmtCheck::new("#")),
+            CheckType::SlashCmt => Box::new(CmtCheck::new("//")),
         })
     }
 }
@@ -650,10 +777,10 @@ pub struct Checker {
 
 impl Checker {
     /// new
-    pub fn new(spec: &CheckSpec, pattern: &str) -> Result<Self> {
+    pub fn new(spec: CheckSpec, pattern: &str) -> Result<Self> {
         Ok(Self {
-            spec: *spec,
             check: spec.make_box(pattern)?,
+            spec,
         })
     }
 }
@@ -688,13 +815,13 @@ impl ColChecker {
         let parts = parts.1.split_once(',');
         if parts.is_none() {
             return err!(
-                "Column Pattern has only one column. SHould be Column,Spec,Pattern : {}",
+                "Column Pattern has only one column. Should be Column,Spec,Pattern : {}",
                 spec
             );
         }
         let parts = parts.unwrap();
         Ok(Self {
-            check: Checker::new(&CheckSpec::new(parts.0)?, parts.1)?,
+            check: Checker::new(CheckSpec::new(parts.0)?, parts.1)?,
             col: nc,
         })
     }
@@ -718,7 +845,7 @@ impl LineCheck for ColChecker {
     }
 }
 
-/// Full list of checkers
+/// Full list of checkers, AND'ed
 #[derive(Default)]
 pub struct CheckList {
     checks: Vec<Box<dyn LineCheck>>,
@@ -764,3 +891,143 @@ impl CheckList {
         true
     }
 }
+/// Full list of checkers, OR'ed
+#[derive(Default)]
+pub struct CheckListOr {
+    checks: Vec<Box<dyn LineCheck>>,
+}
+impl fmt::Debug for CheckListOr {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "CheckListOr")
+    }
+}
+
+impl CheckListOr {
+    /// new
+    pub fn new() -> Self {
+        Self { checks: Vec::new() }
+    }
+    /// is empty?
+    pub fn is_empty(&self) -> bool {
+        self.checks.is_empty()
+    }
+    /// add Check to list
+    pub fn push(&mut self, item: Box<dyn LineCheck>) {
+        self.checks.push(item)
+    }
+    /// lookup
+    pub fn lookup(&mut self, fieldnames: &[&str]) -> Result<()> {
+        for x in self.checks.iter_mut() {
+            x.lookup(fieldnames)?;
+        }
+        Ok(())
+    }
+    /// ok
+    pub fn ok(&self, line: &TextLine) -> bool {
+        for x in &self.checks {
+            if x.check(line) {
+                return true;
+            }
+        }
+        false
+    }
+    /// ok_verbose
+    pub fn ok_verbose(&self, line: &TextLine) -> bool {
+        for x in &self.checks {
+            if x.check(line) {
+                return true;
+            }
+        }
+        false
+    }
+}
+
+/// Full list of checkers, OR'ed
+#[derive(Default)]
+pub struct CheckOr {
+    checks: Vec<Box<dyn BufCheck>>,
+}
+impl fmt::Debug for CheckOr {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "CheckOr")
+    }
+}
+
+impl CheckOr {
+    /// new
+    pub fn new() -> Self {
+        Self { checks: Vec::new() }
+    }
+    /// add Check to list
+    pub fn push(&mut self, item: Box<dyn BufCheck>) {
+        self.checks.push(item)
+    }
+    /// is empty?
+    pub fn is_empty(&self) -> bool {
+        self.checks.is_empty()
+    }
+}
+
+impl BufCheck for CheckOr {
+    fn scheck(&self, buff: &str) -> bool {
+        for x in &self.checks {
+            if x.scheck(buff) {
+                return true;
+            }
+        }
+        false
+    }
+    fn ucheck(&self, buff: &[u8]) -> bool {
+        for x in &self.checks {
+            if x.ucheck(buff) {
+                return true;
+            }
+        }
+        false
+    }
+}
+
+/// Full list of checkers, OR'ed
+#[derive(Default)]
+pub struct CheckAnd {
+    checks: Vec<Box<dyn BufCheck>>,
+}
+impl fmt::Debug for CheckAnd {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "CheckAnd")
+    }
+}
+
+impl CheckAnd {
+    /// new
+    pub fn new() -> Self {
+        Self { checks: Vec::new() }
+    }
+    /// add Check to list
+    pub fn push(&mut self, item: Box<dyn BufCheck>) {
+        self.checks.push(item)
+    }
+    /// is empty?
+    pub fn is_empty(&self) -> bool {
+        self.checks.is_empty()
+    }
+}
+
+impl BufCheck for CheckAnd {
+    fn scheck(&self, buff: &str) -> bool {
+        for x in &self.checks {
+            if !x.scheck(buff) {
+                return false;
+            }
+        }
+        true
+    }
+    fn ucheck(&self, buff: &[u8]) -> bool {
+        for x in &self.checks {
+            if !x.ucheck(buff) {
+                return false;
+            }
+        }
+        true
+    }
+}
