@@ -2,7 +2,8 @@
 //! Also helps with selecting those columns from a line of text
 
 use crate::comp::str_to_u_whole;
-use crate::{err, first, sglob, skip_first, take_first, Error, Result, StringLine, TextLine};
+use crate::text::{Case, Text};
+use crate::{err, Error, Result, StringLine, TextLine};
 use lazy_static::lazy_static;
 use regex::Regex;
 use std::cmp::Ordering;
@@ -10,6 +11,19 @@ use std::collections::HashSet;
 use std::io::Write;
 use std::{fmt, str};
 
+/// What to do if there would be duplicate column names
+#[derive(Debug, Copy, Clone)]
+pub enum DupColHandling {
+    /// Fail if there are duplicate columns names
+    Fail,
+    /// Drop all but the leftmost of the columns
+    Drop,
+    /// Allow duplicate columns names
+    Allow,
+    /// if "foo" is taken, try "foo1", then "foo2" until it's unique
+    Numeric,
+}
+
 /// A column set is a collection of column specifictions, which when interpreted
 /// in the context of a set of named columns, produces a list of column numbers
 /// It contains a bunch of "yes" specs and "no" specs.
@@ -77,7 +91,7 @@ impl OutCol {
         }
     }
     /// new with empty name
-    pub fn from_num(num: usize) -> Self {
+    pub const fn from_num(num: usize) -> Self {
         Self {
             num,
             name: String::new(),
@@ -95,25 +109,25 @@ pub struct ScopedValue {
 impl ScopedValue {
     /// one string
     pub fn new(spec: &str, del: char) -> Self {
-        let mut s = ScopedValue::default();
+        let mut s = Self::default();
         if spec.is_empty() {
             return s;
         }
         let mut spec = spec;
-        let ch = first(spec);
+        let ch = spec.first();
         if ch == del {
-            skip_first(spec);
+            spec.skip_first();
             if spec.is_empty() {
                 s.value.push(del);
                 return s;
             }
-            let ch = first(spec);
+            let ch = spec.first();
             if ch == del {
                 s.value.push(del);
             }
         }
         while !spec.is_empty() {
-            let ch = take_first(&mut spec);
+            let ch = spec.take_first();
             if ch == del {
                 s.cols.add_yes(spec);
                 return s;
@@ -124,7 +138,7 @@ impl ScopedValue {
     }
     /// two strings
     pub fn new2(value: &str, cols: &str) -> Self {
-        let mut s = ScopedValue::default();
+        let mut s = Self::default();
         s.cols.add_yes(cols);
         s.value = value.to_string();
         s
@@ -150,7 +164,7 @@ pub struct ScopedValues {
 impl ScopedValues {
     /// new
     pub fn new() -> Self {
-        ScopedValues::default()
+        Self::default()
     }
     /// value for column not otherwise assigned a value
     pub fn set_default(&mut self, d: &str) {
@@ -239,11 +253,101 @@ pub trait ColumnFun {
     /// write the column names (called once)
     fn write_names(&self, w: &mut dyn Write, head: &StringLine, delim: u8) -> Result<()>;
     /// write the column values (called many times)
-    fn write(&self, w: &mut dyn Write, line: &TextLine, delim: u8) -> Result<()>;
+    fn write(&mut self, w: &mut dyn Write, line: &TextLine, delim: u8) -> Result<()>;
     /// resolve any named columns
     fn lookup(&mut self, fieldnames: &[&str]) -> Result<()>;
 }
 
+/// write the whole line
+#[derive(Debug, Default, Copy, Clone)]
+pub struct ColumnWhole;
+
+impl ColumnFun for ColumnWhole {
+    /// write the column names (called once)
+    fn write_names(&self, w: &mut dyn Write, head: &StringLine, _delim: u8) -> Result<()> {
+        w.write_all(head.line[5..head.line.len() - 1].as_bytes())?;
+        Ok(())
+    }
+    /// write the column values (called many times)
+    fn write(&mut self, w: &mut dyn Write, line: &TextLine, _delim: u8) -> Result<()> {
+        w.write_all(&line.line[0..line.line.len() - 1])?;
+        Ok(())
+    }
+    /// resolve any named columns
+    fn lookup(&mut self, _fieldnames: &[&str]) -> Result<()> {
+        Ok(())
+    }
+}
+
+/// write the whole line
+#[derive(Debug, Default, Clone)]
+pub struct ColumnCount {
+    num: i64,
+    name: String,
+}
+
+impl ColumnCount {
+    /// new
+    pub fn new(num: i64, name: &str) -> Self {
+        Self {
+            num,
+            name: name.to_string(),
+        }
+    }
+}
+
+impl ColumnFun for ColumnCount {
+    /// write the column names (called once)
+    fn write_names(&self, w: &mut dyn Write, _head: &StringLine, _delim: u8) -> Result<()> {
+        w.write_all(self.name.as_bytes())?;
+        Ok(())
+    }
+    /// write the column values (called many times)
+    fn write(&mut self, w: &mut dyn Write, _line: &TextLine, _delim: u8) -> Result<()> {
+        w.write_all(self.num.to_string().as_bytes())?;
+        self.num += 1;
+        Ok(())
+    }
+    /// resolve any named columns
+    fn lookup(&mut self, _fieldnames: &[&str]) -> Result<()> {
+        Ok(())
+    }
+}
+
+/// write the whole line
+#[derive(Debug, Default, Clone)]
+pub struct ColumnLiteral {
+    value: Vec<u8>,
+    name: String,
+}
+
+impl ColumnLiteral {
+    /// new
+    pub fn new(value: &[u8], name: &str) -> Self {
+        Self {
+            value: value.to_vec(),
+            name: name.to_string(),
+        }
+    }
+}
+
+impl ColumnFun for ColumnLiteral {
+    /// write the column names (called once)
+    fn write_names(&self, w: &mut dyn Write, _head: &StringLine, _delim: u8) -> Result<()> {
+        w.write_all(self.name.as_bytes())?;
+        Ok(())
+    }
+    /// write the column values (called many times)
+    fn write(&mut self, w: &mut dyn Write, _line: &TextLine, _delim: u8) -> Result<()> {
+        w.write_all(&self.value)?;
+        Ok(())
+    }
+    /// resolve any named columns
+    fn lookup(&mut self, _fieldnames: &[&str]) -> Result<()> {
+        Ok(())
+    }
+}
+
 impl fmt::Debug for dyn ColumnFun {
     fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
         write!(f, "ColumnFun")
@@ -253,11 +357,16 @@ impl fmt::Debug for dyn ColumnFun {
 impl ColumnSet {
     /// Create an empty column set, which selects all columns.
     pub fn new() -> Self {
-        ColumnSet::default()
+        Self::default()
     }
+    /// is empty?
+    pub fn is_empty(&self) -> bool {
+        self.pos.is_empty() && self.neg.is_empty()
+    }
+
     /// Create an column set from a spec, e.g. "1-3"
     pub fn from_spec(spec: &str) -> Self {
-        let mut s = ColumnSet::default();
+        let mut s = Self::default();
         s.add_yes(spec);
         s
     }
@@ -379,7 +488,7 @@ impl ColumnSet {
     fn glob_range(fieldnames: &[&str], rng: &str) -> Result<Vec<usize>> {
         let mut ret = Vec::new();
         for f in fieldnames.iter().enumerate() {
-            if sglob(rng, f.1, false) {
+            if f.1.glob(rng, Case::Sens) {
                 // allow case insensitive?
                 ret.push(f.0);
             }
@@ -435,7 +544,7 @@ impl ColumnSet {
 
     /// turn comma delimited list of ranges into list of possibly named column numbers
     pub fn ranges(fieldnames: &[&str], rng: &str) -> Result<Vec<OutCol>> {
-        let mut c = ColumnSet::new();
+        let mut c = Self::new();
         c.add_yes(rng);
         c.lookup(fieldnames)?;
         Ok(c.get_cols_full())
@@ -473,17 +582,16 @@ impl ColumnSet {
         }
 
         let start: usize;
-        let end: usize;
-        if parts.len() == 1 {
+        let end = if parts.len() == 1 {
             start = Self::single(fieldnames, parts[0])?;
-            end = start;
+            start
         } else {
             if parts[1].is_empty() {
                 parts[1] = "+1";
             }
             start = Self::single(fieldnames, parts[0])?;
-            end = Self::single(fieldnames, parts[1])?;
-        }
+            Self::single(fieldnames, parts[1])?
+        };
 
         if start > end {
             // throw new CdxException("start > end, i.e. {" + parts.get(0) + " > " + parts.get(1));
@@ -708,11 +816,12 @@ impl ColumnSet {
     }
 
     /// return owned columns by const reference
-    pub fn get_cols(&self) -> &Vec<OutCol> {
+    pub const fn get_cols(&self) -> &Vec<OutCol> {
         &self.columns
     }
 
     /// steal owned columns by value
+    #[allow(clippy::missing_const_for_fn)] // clippy bug
     pub fn get_cols_full(self) -> Vec<OutCol> {
         self.columns
     }
@@ -735,7 +844,7 @@ impl ColumnSet {
     ///    assert_eq!(ColumnSet::lookup_cols("~2-3", &header).unwrap(), &[0,3,4]);
     /// ```
     pub fn lookup_cols(spec: &str, names: &[&str]) -> Result<Vec<usize>> {
-        let mut s = ColumnSet::new();
+        let mut s = Self::new();
         s.add_yes(spec);
         s.lookup(names)?;
         Ok(s.get_cols_num())
@@ -743,7 +852,7 @@ impl ColumnSet {
 
     /// Shorthand to look up some columns, with names
     pub fn lookup_cols_full(spec: &str, names: &[&str]) -> Result<Vec<OutCol>> {
-        let mut s = ColumnSet::new();
+        let mut s = Self::new();
         s.add_yes(spec);
         s.lookup(names)?;
         Ok(s.get_cols_full())
@@ -759,7 +868,7 @@ impl ColumnSet {
     ///    assert_eq!(ColumnSet::lookup1("three", &header).unwrap(), 3);
     /// ```
     pub fn lookup1(spec: &str, names: &[&str]) -> Result<usize> {
-        let mut s = ColumnSet::new();
+        let mut s = Self::new();
         s.add_yes(spec);
         s.lookup(names)?;
         if s.get_cols().len() != 1 {
@@ -773,49 +882,21 @@ impl ColumnSet {
     }
 }
 
-/*
-impl ColumnFun for ReaderFull<'_> {
-    fn write(&self, w: &mut dyn Write, line: &TextLine, delim: u8) -> Result<()> {
-        let mut need_delim = false;
-        for x in line.iter() {
-            if need_delim {
-                w.write_all(&[delim])?;
-            }
-            need_delim = true;
-            w.write_all(x)?;
-        }
-        Ok(())
-    }
-
-    fn write_names(&self, w: &mut dyn Write, head: &StringLine, delim: u8) -> Result<()> {
-        let mut need_delim = false;
-        for x in head.iter() {
-            if need_delim {
-                w.write_all(&[delim])?;
-            }
-            need_delim = true;
-            w.write_all(x)?;
-        }
-        Ok(())
-    }
-}
- */
-
 /// Write the columns with a custom delimiter
-pub struct ColumnClump<'a> {
-    cols: Box<dyn ColumnFun + 'a>,
+pub struct ColumnClump {
+    cols: Box<dyn ColumnFun>,
     name: String,
     delim: u8,
 }
-impl fmt::Debug for ColumnClump<'_> {
+impl fmt::Debug for ColumnClump {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "ColumnClump")
     }
 }
 
-impl<'a> ColumnClump<'a> {
+impl ColumnClump {
     /// new ColumnClump from parts
-    pub fn new(cols: Box<dyn ColumnFun + 'a>, name: &str, delim: u8) -> Self {
+    pub fn new(cols: Box<dyn ColumnFun>, name: &str, delim: u8) -> Self {
         Self {
             cols,
             name: name.to_string(),
@@ -829,7 +910,7 @@ impl<'a> ColumnClump<'a> {
         if spec.is_empty() {
             return err!("Can't construct a group from an empty string");
         }
-        let delim = take_first(&mut spec);
+        let delim = spec.take_first();
         if delim.len_utf8() != 1 {
             return err!("Delimiter must be a plain ascii character : {}", orig_spec);
         }
@@ -852,8 +933,8 @@ impl<'a> ColumnClump<'a> {
     }
 }
 
-impl<'a> ColumnFun for ColumnClump<'a> {
-    fn write(&self, w: &mut dyn Write, line: &TextLine, _delim: u8) -> Result<()> {
+impl ColumnFun for ColumnClump {
+    fn write(&mut self, w: &mut dyn Write, line: &TextLine, _delim: u8) -> Result<()> {
         self.cols.write(w, line, self.delim)?;
         Ok(())
     }
@@ -875,7 +956,7 @@ pub struct ReaderColumns {
 
 impl ReaderColumns {
     /// new ReaderColumns
-    pub fn new(columns: ColumnSet) -> Self {
+    pub const fn new(columns: ColumnSet) -> Self {
         Self { columns }
     }
 }
@@ -891,7 +972,7 @@ pub fn write_colname(w: &mut dyn Write, col: &OutCol, head: &StringLine) -> Resu
 }
 
 impl ColumnFun for ReaderColumns {
-    fn write(&self, w: &mut dyn Write, line: &TextLine, delim: u8) -> Result<()> {
+    fn write(&mut self, w: &mut dyn Write, line: &TextLine, delim: u8) -> Result<()> {
         self.columns.write3(w, line, delim)?;
         Ok(())
     }
@@ -917,17 +998,17 @@ impl ColumnFun for ReaderColumns {
 }
 
 /// A collection of ColumnFun writers
-pub struct Writer<'a> {
-    v: Vec<Box<dyn ColumnFun + 'a>>,
+pub struct Writer {
+    v: Vec<Box<dyn ColumnFun>>,
     delim: u8,
 }
-impl fmt::Debug for Writer<'_> {
+impl fmt::Debug for Writer {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "Writer")
     }
 }
 
-impl<'a> Writer<'a> {
+impl Writer {
     /// new Writer
     pub fn new(delim: u8) -> Self {
         Self {
@@ -947,7 +1028,7 @@ impl<'a> Writer<'a> {
         Ok(())
     }
     /// Add a new writer
-    pub fn push(&mut self, x: Box<dyn ColumnFun + 'a>) {
+    pub fn push(&mut self, x: Box<dyn ColumnFun>) {
         self.v.push(x);
     }
     /// Write the column names
@@ -961,8 +1042,8 @@ impl<'a> Writer<'a> {
         Ok(())
     }
     /// Write the column values
-    pub fn write(&self, w: &mut dyn Write, line: &TextLine) -> Result<()> {
-        let mut iter = self.v.iter();
+    pub fn write(&mut self, w: &mut dyn Write, line: &TextLine) -> Result<()> {
+        let mut iter = self.v.iter_mut();
         match iter.next() {
             None => {}
             Some(first) => {
@@ -978,7 +1059,7 @@ impl<'a> Writer<'a> {
     }
 }
 
-impl<'a> Default for Writer<'a> {
+impl Default for Writer {
     fn default() -> Self {
         Self::new(b'\t')
     }
@@ -997,7 +1078,7 @@ pub struct NamedCol {
 
 impl NamedCol {
     /// new named column
-    pub fn new() -> Self {
+    pub const fn new() -> Self {
         Self {
             name: String::new(),
             num: 0,
@@ -1028,11 +1109,11 @@ impl NamedCol {
         if spec.is_empty() {
             return Ok(spec);
         }
-        let mut ch = first(spec);
+        let mut ch = spec.first();
         let was_plus = ch == '+';
         if was_plus {
-            spec = skip_first(spec);
-            ch = first(spec);
+            spec = spec.skip_first();
+            ch = spec.first();
         }
         if ch.is_ascii_digit() {
             let mut pos: usize = 0;
@@ -1103,7 +1184,7 @@ pub struct CompositeColumn {
 impl CompositeColumn {
     /// new
     pub fn new(s: &str) -> Result<Self> {
-        let mut c = CompositeColumn::default();
+        let mut c = Self::default();
         c.set(s)?;
         Ok(c)
     }
@@ -1136,23 +1217,23 @@ impl CompositeColumn {
         self.parts.clear();
         const TAG: char = '^';
         while !curr.is_empty() {
-            let ch = take_first(&mut curr);
+            let ch = curr.take_first();
             if ch == TAG {
                 if curr.is_empty() {
                     self.suffix.push(TAG);
-                } else if first(curr) == TAG {
+                } else if curr.first() == TAG {
                     self.suffix.push(TAG);
-                    curr = skip_first(curr);
+                    curr = curr.skip_first();
                 } else {
                     let mut p = CompositeColumnPart::new();
                     std::mem::swap(&mut p.prefix, &mut self.suffix);
-                    if first(curr) == '{' {
-                        curr = skip_first(curr);
+                    if curr.first() == '{' {
+                        curr = curr.skip_first();
                         curr = p.col.parse(curr)?;
-                        if curr.is_empty() || first(curr) != '}' {
+                        if curr.is_empty() || curr.first() != '}' {
                             return err!("Closing bracket not found : {}", spec);
                         }
-                        curr = skip_first(curr);
+                        curr = curr.skip_first();
                     } else {
                         curr = p.col.parse(curr)?;
                     }
@@ -1174,7 +1255,7 @@ impl ColumnFun for CompositeColumn {
         Ok(())
     }
     /// write the column values (called many times)
-    fn write(&self, w: &mut dyn Write, line: &TextLine, _delim: u8) -> Result<()> {
+    fn write(&mut self, w: &mut dyn Write, line: &TextLine, _delim: u8) -> Result<()> {
         for x in &self.parts {
             w.write_all(x.prefix.as_bytes())?;
             w.write_all(line.get(x.col.num))?;
@@ -1184,7 +1265,7 @@ impl ColumnFun for CompositeColumn {
     }
     /// resolve any named columns
     fn lookup(&mut self, fieldnames: &[&str]) -> Result<()> {
-        CompositeColumn::lookup(self, fieldnames)
+        Self::lookup(self, fieldnames)
     }
 }
 
