@@ -3,7 +3,6 @@
 //! It is hoped that the associated library will be useful for theid party tools.
 
 #![warn(
-    missing_copy_implementations,
     absolute_paths_not_starting_with_crate,
     explicit_outlives_requirements,
     keyword_idents,
@@ -20,25 +19,39 @@
     unreachable_pub,
     unused_lifetimes,
     unused_extern_crates,
-    unused_qualifications
+    unused_qualifications,
+
+//    clippy::all,
+//    clippy::restriction,
+//    clippy::pedantic,
+    clippy::nursery,
+    clippy::cargo,
 )]
+#![allow(clippy::multiple_crate_versions)]
 
 use flate2::read::MultiGzDecoder;
 use fs_err as fs;
 //use std::os::unix::io::FromRawFd;
 use std::error;
+use std::ffi::OsStr;
 use std::io::{self, BufRead, Read, Write};
 use std::ops::{Deref, DerefMut};
+use std::os::unix::ffi::OsStrExt;
+use std::path::Path;
 use std::{fmt, str};
 
+pub mod binsearch;
 pub mod check;
 pub mod column;
 pub mod comp;
 pub mod expr;
 pub mod join;
 pub mod sort;
+pub mod text;
 pub mod tooltest;
 
+use crate::text::Text;
+
 /// Shorthand for returning an error Result
 #[macro_export]
 macro_rules! err {
@@ -95,7 +108,7 @@ impl Error {
         }
     }
     /// return true if this error should be treated as an error, bu silently
-    pub fn silent(&self) -> bool {
+    pub const fn silent(&self) -> bool {
         matches!(self, Error::Silent)
     }
 }
@@ -199,16 +212,27 @@ impl std::ops::Index<usize> for StringLine {
     }
 }
 
+/// write a buffer. If non-empty, ensure trailing newline
+pub fn write_all_nl(w: &mut impl Write, buf: &[u8]) -> Result<()> {
+    if !buf.is_empty() {
+        w.write_all(buf)?;
+        if buf[buf.len() - 1] != b'\n' {
+            w.write_all(&[b'\n'])?;
+        }
+    }
+    Ok(())
+}
+
 impl TextLine {
     /// make a new TextLine
-    pub fn new() -> Self {
+    pub const fn new() -> Self {
         Self {
             line: Vec::new(),
             parts: Vec::new(),
         }
     }
     /// Iterator over columns in the line
-    pub fn iter(&self) -> TextLineIter<'_> {
+    pub const fn iter(&self) -> TextLineIter<'_> {
         TextLineIter {
             line: self,
             index: 0,
@@ -280,19 +304,29 @@ impl TextLine {
 
 impl StringLine {
     /// make a new StringLine
-    pub fn new() -> Self {
+    pub const fn new() -> Self {
         Self {
             line: String::new(),
             parts: Vec::new(),
         }
     }
     /// Iterator over columns in the line
-    pub fn iter(&self) -> StringLineIter<'_> {
+    pub const fn iter(&self) -> StringLineIter<'_> {
         StringLineIter {
             line: self,
             index: 0,
         }
     }
+    /// create a fake CDX header with columns c1,c2...
+    pub fn fake(&mut self, num_cols: usize, delim: u8) {
+        self.line = " CDX".to_string();
+        for i in 1..=num_cols {
+            self.line += std::str::from_utf8(&[delim]).unwrap();
+            self.line += "c";
+            self.line += &i.to_string();
+        }
+        self.line += "\n";
+    }
     fn clear(&mut self) {
         self.parts.clear();
         self.line.clear();
@@ -414,7 +448,7 @@ impl<'a> Iterator for StringLineIter<'a> {
     type Item = &'a str;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.index >= self.line.len() {
+        if self.index >= self.line.parts.len() {
             None
         } else {
             self.index += 1;
@@ -462,8 +496,35 @@ impl DerefMut for Infile {
     }
 }
 
+impl AsRef<io::BufReader<Box<dyn Read>>> for Infile {
+    fn as_ref(&self) -> &io::BufReader<Box<dyn Read>> {
+        &self.0
+    }
+}
+
+impl AsMut<io::BufReader<Box<dyn Read>>> for Infile {
+    fn as_mut(&mut self) -> &mut io::BufReader<Box<dyn Read>> {
+        &mut self.0
+    }
+}
+
 type Outfile = io::BufWriter<Box<dyn Write>>;
 
+/// Make an Outfile from a file name
+pub fn get_writer2<P: AsRef<Path>>(name: P) -> Result<Outfile> {
+    let name = name.as_ref().as_os_str();
+    let inner: Box<dyn Write> = {
+        if name == OsStr::new("-") {
+            Box::new(io::stdout())
+        } else if name == OsStr::new("--") {
+            Box::new(io::stderr())
+        } else {
+            Box::new(fs::OpenOptions::new().write(true).create(true).open(name)?)
+        }
+    };
+    Ok(io::BufWriter::new(inner))
+}
+
 /// Make an Outfile from a file name
 pub fn get_writer(name: &str) -> Result<Outfile> {
     let inner: Box<dyn Write> = {
@@ -478,6 +539,7 @@ pub fn get_writer(name: &str) -> Result<Outfile> {
     Ok(io::BufWriter::new(inner))
 }
 
+// should return Cow<>
 fn unescape_vec(data: &[u8]) -> Vec<u8> {
     let mut ret: Vec<u8> = Vec::with_capacity(data.len());
     let mut last_was_slash = false;
@@ -502,6 +564,29 @@ fn unescape_vec(data: &[u8]) -> Vec<u8> {
     ret
 }
 
+/// Make an Infile from a file name
+pub fn get_reader2<P: AsRef<Path>>(name: P) -> Result<Infile> {
+    let name = name.as_ref().as_os_str();
+    let inner: Box<dyn Read> = {
+        if name == OsStr::new("-") {
+            //	    unsafe { Box::new(std::fs::File::from_raw_fd(1)) }
+            Box::new(io::stdin())
+            //            } else if name.starts_with("s3://") {
+            //                Box::new(open_s3_file(name)?)
+        } else if name.as_bytes().starts_with(b"<<") {
+            Box::new(std::io::Cursor::new(unescape_vec(&name.as_bytes()[2..])))
+        } else {
+            Box::new(fs::File::open(name)?)
+        }
+    };
+    let mut outer = io::BufReader::new(inner);
+    let start = outer.fill_buf()?;
+    if start.starts_with(&[0x1fu8, 0x8bu8, 0x08u8]) {
+        outer = io::BufReader::new(Box::new(MultiGzDecoder::new(outer)));
+    }
+    Ok(Infile::new(outer))
+}
+
 /// Make an Infile from a file name
 pub fn get_reader(name: &str) -> Result<Infile> {
     let inner: Box<dyn Read> = {
@@ -525,23 +610,31 @@ pub fn get_reader(name: &str) -> Result<Infile> {
 }
 
 #[derive(Debug, Default)]
-struct InfileContext {
-    header: StringLine,
-    delim: u8,
-    is_done: bool,
-    is_empty: bool,
+/// shared context for any input file type
+pub struct InfileContext {
+    /// CDX header, contructed if necessary
+    pub header: StringLine,
+    /// delimter
+    pub delim: u8,
+    /// have we read all the btes of the file
+    pub is_done: bool,
+    /// is the file length zero
+    pub is_empty: bool,
+    /// was there a CDX header?
+    pub has_header: bool,
 }
 
 // FIXME -- specify delimiter
 // if CDX and specified and different, then strip header
 /// Reader header line, if any, and first line of text
 impl InfileContext {
-    fn new() -> Self {
+    const fn new() -> Self {
         Self {
             header: StringLine::new(),
             delim: b'\t',
             is_done: true,
             is_empty: true,
+            has_header: false,
         }
     }
     fn read_header(&mut self, file: &mut impl BufRead, line: &mut TextLine) -> Result<()> {
@@ -553,6 +646,7 @@ impl InfileContext {
         self.is_done = false;
         self.is_empty = false;
         if self.header.line.starts_with(" CDX") {
+            self.has_header = true;
             self.delim = self.header.line.as_bytes()[4];
             self.header.split(self.delim);
             self.header.parts.remove(0);
@@ -576,7 +670,6 @@ impl InfileContext {
             let mut fake_head = head_str.as_bytes();
             self.header.read(&mut fake_head)?;
             self.header.split(self.delim);
-            self.header.line.clear();
         }
         Ok(())
     }
@@ -588,7 +681,8 @@ pub struct Reader {
     file: Infile,
     /// The current line of text
     pub line: TextLine,
-    cont: InfileContext,
+    /// context
+    pub cont: InfileContext,
     /// automatically split each TextLine
     pub do_split: bool,
 }
@@ -603,12 +697,27 @@ impl Reader {
             do_split: true,
         }
     }
+    /// make a new Reader
+    pub fn new_open(name: &str) -> Result<Self> {
+        let mut tmp = Self {
+            file: get_reader(name)?,
+            line: TextLine::new(),
+            cont: InfileContext::new(),
+            do_split: true,
+        };
+        tmp.cont.read_header(&mut *tmp.file, &mut tmp.line)?;
+        Ok(tmp)
+    }
     /// get current line
-    pub fn curr(&self) -> &TextLine {
+    pub const fn curr(&self) -> &TextLine {
         &self.line
     }
+    /// get current line contents, without the trailing newline
+    pub fn curr_nl(&self) -> &[u8] {
+        &self.line.line[0..self.line.line.len() - 1]
+    }
     /// get header
-    pub fn header(&self) -> &StringLine {
+    pub const fn header(&self) -> &StringLine {
         &self.cont.header
     }
     /// get column names
@@ -621,11 +730,11 @@ impl Reader {
         self.cont.read_header(&mut *self.file, &mut self.line)
     }
     /// was the file zero bytes?
-    pub fn is_empty(&self) -> bool {
+    pub const fn is_empty(&self) -> bool {
         self.cont.is_empty
     }
     /// have we read all the lines?
-    pub fn is_done(&self) -> bool {
+    pub const fn is_done(&self) -> bool {
         self.cont.is_done
     }
     /// write the current text line with newline
@@ -633,6 +742,11 @@ impl Reader {
         w.write_all(&self.line.line)?;
         Ok(())
     }
+    /// write header
+    pub fn write_header(&self, w: &mut impl Write) -> Result<()> {
+        w.write_all(self.cont.header.line.as_bytes())?;
+        Ok(())
+    }
     /// get the next line of text. Return true if no more data available.
     pub fn getline(&mut self) -> Result<bool> {
         if !self.cont.is_done {
@@ -652,7 +766,8 @@ impl Reader {
 pub struct LookbackReader {
     file: Infile,
     lines: Vec<TextLine>,
-    cont: InfileContext,
+    /// context
+    pub cont: InfileContext,
     /// Automatically split each line into columns
     pub do_split: bool,
     curr: usize,
@@ -681,15 +796,15 @@ impl LookbackReader {
         self.cont.read_header(&mut *self.file, &mut self.lines[0])
     }
     /// The full text of the header, without the trailing newline
-    pub fn header(&self) -> &String {
+    pub const fn header(&self) -> &String {
         &self.cont.header.line
     }
     /// was file zero bytes?
-    pub fn is_empty(&self) -> bool {
+    pub const fn is_empty(&self) -> bool {
         self.cont.is_empty
     }
     /// have we hit EOF?
-    pub fn is_done(&self) -> bool {
+    pub const fn is_done(&self) -> bool {
         self.cont.is_done
     }
     fn incr(&mut self) {
@@ -733,14 +848,13 @@ impl LookbackReader {
     }
 }
 
-fn prerr(data: &[&[u8]]) -> Result<()> {
+fn prerr(data: &[&[u8]]) {
     for x in data {
-        std::io::stderr().write_all(x)?;
+        std::io::stderr().write_all(x).unwrap();
     }
-    std::io::stderr().write_all(b"\n")?;
-    Ok(())
+    std::io::stderr().write_all(b"\n").unwrap();
 }
-
+/*
 fn bytes_equal(c1: u8, c2: u8, ic: bool) -> bool {
     if c1 == c2 {
         return true;
@@ -766,7 +880,8 @@ fn chars_equal(c1: char, c2: char, ic: bool) -> bool {
     }
     c1.to_lowercase().eq(c2.to_lowercase())
 }
-
+*/
+/*
 /// match in glob format
 pub fn bglob(mut wild: &[u8], mut buff: &[u8], ic: bool) -> bool {
     while !buff.is_empty() && !wild.is_empty() && (wild[0] != b'*') {
@@ -811,11 +926,11 @@ fn first(s: &str) -> char {
     debug_assert!(!s.is_empty());
     s.chars().next().unwrap()
 }
-
+*/
 //fn first_len(s : &str) -> usize {
 //    s.chars().next().unwrap().len_utf8()
 //}
-
+/*
 fn skip_first(s: &str) -> &str {
     debug_assert!(!s.is_empty());
     &s[s.chars().next().unwrap().len_utf8()..]
@@ -867,7 +982,7 @@ pub fn sglob(mut wild: &str, mut buff: &str, ic: bool) -> bool {
     }
     wild.is_empty()
 }
-
+*/
 /// How to combine headers from multiple sources
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum HeaderMode {
@@ -891,17 +1006,17 @@ impl str::FromStr for HeaderMode {
     type Err = Error;
     fn from_str(spec: &str) -> Result<Self> {
         if spec.to_ascii_lowercase() == "match" {
-            Ok(HeaderMode::Match)
+            Ok(Self::Match)
         } else if spec.to_ascii_lowercase() == "require" {
-            Ok(HeaderMode::Require)
+            Ok(Self::Require)
         } else if spec.to_ascii_lowercase() == "strip" {
-            Ok(HeaderMode::Strip)
+            Ok(Self::Strip)
         } else if spec.to_ascii_lowercase() == "none" {
-            Ok(HeaderMode::None)
+            Ok(Self::None)
         } else if spec.to_ascii_lowercase() == "trust" {
-            Ok(HeaderMode::Trust)
+            Ok(Self::Trust)
         } else if spec.to_ascii_lowercase() == "ignore" {
-            Ok(HeaderMode::Ignore)
+            Ok(Self::Ignore)
         } else {
             err!(
                 "Input Header Mode must be one of Match, Require, Strip, None or Trust : {}",
@@ -913,7 +1028,7 @@ impl str::FromStr for HeaderMode {
 
 impl Default for HeaderMode {
     fn default() -> Self {
-        HeaderMode::Match
+        Self::Match
     }
 }
 
@@ -961,7 +1076,7 @@ fn is_valid_cdx(data_in: &[u8], mode: HeaderMode, fname: &str) -> Result<bool> {
         if x.is_empty() {
             return err!("File {} has an empty column name", fname);
         }
-        if !first(x).is_alphabetic() {
+        if !x.first().is_alphabetic() {
             return err!("Header for file {} has column name {} which does not start with an alphabetic character.", fname, x);
         }
         for ch in x.chars() {
@@ -985,6 +1100,18 @@ impl HeaderChecker {
             ..Self::default()
         }
     }
+    /// call for the first line of every input file
+    /// return true if the header should be written
+    pub fn check_file(&mut self, file: &InfileContext, fname: &str) -> Result<bool> {
+        let first = !self.saw_one;
+        if file.has_header {
+            self.check(file.header.line.as_bytes(), fname)?;
+        } else {
+            self.check(b"fake", fname)?;
+        }
+        Ok(file.has_header && first)
+    }
+
     /// call for the first line of every input file
     /// return true if the line should be part of the output
     pub fn check(&mut self, first_line: &[u8], fname: &str) -> Result<bool> {
@@ -1054,6 +1181,19 @@ impl HeaderChecker {
     }
 }
 
+/// copy r to w
+pub fn copy(mut r: impl Read, mut w: impl Write) -> Result<()> {
+    let mut buff = [0u8; 16 * 1024];
+    loop {
+        let sz = r.read(&mut buff)?;
+        if sz == 0 {
+            return Ok(());
+        }
+        w.write_all(&buff[0..sz])?;
+    }
+}
+
+/*
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -1088,12 +1228,6 @@ mod tests {
         assert!(bglob(b"a?b", b"anb", false));
         assert!(!bglob(b"a?b", "añb".as_bytes(), false));
 
-        assert!(bglob(
-	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/*/add_hour_*",
-	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
-        assert!(bglob(
-	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/???/add_hour_*",
-	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
     }
     // ñ
     #[test]
@@ -1131,11 +1265,6 @@ mod tests {
         assert!(sglob("añb", "aÑb", true));
         assert!(sglob("aÑb", "añb", true));
 
-        assert!(sglob(
-	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/*/add_hour_*",
-	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
-        assert!(sglob(
-	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/???/add_hour_*",
-	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
     }
 }
+*/
