@@ -2,11 +2,14 @@
 use crate::comp::{skip_leading_white, str_to_i_whole};
 use crate::prerr;
 use crate::{err, get_reader, get_writer, Error, Infile, Result};
+//use fs_err as fs;
 use fs_err as fs;
 use std::ffi::OsStr;
-use std::io::{BufRead, Write};
+use std::io::{BufRead, Read, Write};
 use std::os::unix::ffi::OsStrExt;
+use std::path::Path;
 use std::process::Command;
+use tempdir::TempDir;
 
 #[derive(Debug, Clone, Default)]
 struct OneFile {
@@ -27,6 +30,21 @@ pub struct Test {
     out_files: Vec<OneFile>,
 }
 
+/// remove item from Vec
+fn nuke(v: &mut Vec<String>, s: &str) {
+    v.retain(|x| x != s);
+}
+
+/// return directory as vec of file names
+pub fn read_dir(dir: &Path) -> Result<Vec<String>> {
+    let mut dirs = Vec::new();
+    for entry in fs::read_dir(dir)? {
+        let entry = entry?;
+        dirs.push(entry.file_name().to_string_lossy().to_string());
+    }
+    Ok(dirs)
+}
+
 /// if line doesn't start with tag, return false
 /// read following lines into buff, until another # line is seen
 /// set need_read to false if line ends up with unexamined data
@@ -39,7 +57,7 @@ fn grab(
 ) -> Result<bool> {
     if let Some(x) = line.strip_prefix(tag) {
         if !x.is_empty() {
-            prerr(&[b"Unexpected stuff after ", tag, b" : ", line])?;
+            prerr(&[b"Unexpected stuff after ", tag, b" : ", line]);
             return Err(Error::Silent);
         }
         loop {
@@ -148,18 +166,15 @@ impl Test {
 
     /// run the test
     pub fn run(&mut self, config: &Config) -> Result<bool> {
-        let mut tmp = std::env::temp_dir();
-        tmp.push("toottest");
-        let _ = fs::remove_dir_all(&tmp);
-        fs::create_dir(&tmp)?;
+        let tmp = TempDir::new("tooltest")?;
         for x in &self.in_files {
-            let mut fname = tmp.clone();
+            let mut fname = tmp.path().to_owned();
             fname.push(&x.name);
             let mut w = get_writer(&fname.to_string_lossy())?;
             w.write_all(&x.content)?;
         }
         #[allow(clippy::redundant_clone)]
-        let mut tmp_stdin = tmp.clone();
+        let mut tmp_stdin = tmp.path().to_owned();
         tmp_stdin.push("stdin");
         {
             let mut w = std::fs::OpenOptions::new()
@@ -169,10 +184,10 @@ impl Test {
             w.write_all(&self.stdin)?;
         }
         let ncmd = String::from_utf8_lossy(&self.cmd)
-            .replace("$TMP", &tmp.as_path().display().to_string())
+            .replace("$TMP", &tmp.path().display().to_string())
             .as_bytes()
             .to_vec();
-        //        prerr(&[b"About to run ", &ncmd])?;
+        //        prerr(&[b"About to run ", &ncmd]);
         let cmd: Vec<&[u8]> = ncmd.split(|num| num <= &b' ').collect();
         if cmd.is_empty() {
             return err!("command is empty");
@@ -181,7 +196,7 @@ impl Test {
         basecmd.extend(cmd[0]);
         let mut tmp_cmd = Command::new(OsStr::from_bytes(&basecmd));
         for x in &cmd[1..] {
-            //	    prerr(&[b"with arg ", x])?;
+            //	    prerr(&[b"with arg ", x]);
             tmp_cmd.arg(OsStr::from_bytes(x));
         }
         let res = tmp_cmd
@@ -190,14 +205,18 @@ impl Test {
         let output;
         match res {
             Err(x) => {
-                prerr(&[b"Error trying to execute : ", &basecmd])?;
+                prerr(&[b"Error trying to execute : ", &basecmd]);
                 return Err(Error::IoError(x));
             }
             Ok(x) => {
                 output = x;
             }
         }
-        std::fs::remove_file(&tmp_stdin)?;
+        let mut files = read_dir(tmp.path())?;
+        nuke(&mut files, &"stdin".to_string());
+        for x in &self.in_files {
+            nuke(&mut files, &x.name);
+        }
         let mut failed = false;
         match output.status.code() {
             Some(code) => {
@@ -220,7 +239,7 @@ impl Test {
                 &output.stderr,
                 b"\ninstead of\n",
                 &self.stderr,
-            ])?;
+            ]);
         }
         if output.stdout != self.stdout {
             failed = true;
@@ -229,11 +248,45 @@ impl Test {
                 &output.stdout,
                 b"\ninstead of\n",
                 &self.stdout,
-            ])?;
+            ]);
+        }
+        for x in &self.out_files {
+            nuke(&mut files, &x.name);
+            let mut fname = tmp.path().to_owned();
+            fname.push(&x.name);
+            match get_reader(&fname.to_string_lossy()) {
+                Err(e) => {
+                    failed = true;
+                    eprintln!("{:?}", e);
+                }
+                Ok(mut f) => {
+                    let mut body = Vec::new();
+                    f.read_to_end(&mut body)?;
+                    if body != x.content {
+                        failed = true;
+                        prerr(&[
+                            b"File ",
+                            x.name.as_bytes(),
+                            b" was\n",
+                            &body,
+                            b" but should have been\n",
+                            &x.content,
+                        ]);
+                    }
+                }
+            }
+        }
+        if !files.is_empty() {
+            failed = true;
+            eprintln!("Files left over in TMP dir :");
+            for x in &files {
+                eprintln!("{}", x);
+            }
         }
         if failed {
-            prerr(&[b"Test ", self.name.as_bytes(), b" failed ", &self.cmd])?;
+            prerr(&[b"Test ", self.name.as_bytes(), b" failed ", &self.cmd]);
         }
+        // if (Debug) tmp.into_path()
         Ok(!failed)
     }
 }
